/**
 * DREAM Copyright (C) 2003 2004 INRIA Rhone-Alpes
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option) any
 * later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * Contact : dream@objectweb.org
 * 
 * Initial developer(s): Matthieu Leclercq Contributor(s):
 */

package org.objectweb.dream.channel;

import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectInputStream;
import java.io.ObjectOutput;
import java.io.ObjectOutputStream;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;

import org.objectweb.dream.InitializationException;
import org.objectweb.dream.PushException;
import org.objectweb.dream.control.activity.Util;
import org.objectweb.dream.control.activity.task.AbstractTask;
import org.objectweb.dream.control.activity.task.Task;
import org.objectweb.dream.control.activity.task.TaskController;
import org.objectweb.dream.control.activity.task.thread.ThreadPoolAttributeController;
import org.objectweb.dream.control.activity.task.thread.ThreadPoolController;
import org.objectweb.dream.control.activity.task.thread.ThreadPoolOverflowException;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.util.Error;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.Interface;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.julia.control.lifecycle.ChainedIllegalLifeCycleException;
import org.objectweb.fractal.util.Fractal;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * ChannelIn implementation accepting multiple connection at the same time. For
 * each connection a thread is created to read incomming messages.
 */
public class MultiConnTCPChannelInImpl extends AbstractTCPChannelInImpl
    implements
      MultiConnTCPChannelInAttributeController
{

  /** The {@link ServerSocket}used by this component */
  protected ServerSocket                  listen;

  protected Task                          connectionTask      = new ConnectionTask();
  protected Task                          inTask              = new InTask();

  protected ThreadPoolController          threadPoolController;
  protected ThreadPoolAttributeController threadPoolAttributeController;

  protected LinkedList                    availableSocketList = new LinkedList();
  protected int                           maxConnection;
  protected int                           nbConnection        = 0;

  protected int                           soTimeout           = -1;

  // ---------------------------------------------------------------------------
  // Inner task class implementation
  // ---------------------------------------------------------------------------

  protected class ConnectionTask extends AbstractTask
  {

    /**
     * Constructor
     */
    public ConnectionTask()
    {
      super("ChannelIn-connection-task");
    }

    /**
     * @see Task#execute(Object)
     */
    public Object execute(Object hints) throws InterruptedException
    {
      Socket socket;
      if (listen == null || listen.isClosed())
      {
        logger.log(BasicLevel.DEBUG,
            "ChannelIn-connection-task : ServerSocket closed, create new one");
        try
        {
          listen = createServerSocket(portNum);
        }
        catch (IOException e)
        {
          logger.log(BasicLevel.ERROR,
              "ChannelIn-connection-task : unable to open server socket", e);
          return STOP_EXECUTING;
        }
      }
      try
      {
        // connect
        logger.log(BasicLevel.DEBUG,
            "ChannelIn-connection-task : waiting connection");
        socket = listen.accept();
      }
      catch (SocketException e)
      {
        logger.log(BasicLevel.WARN, e);
        return EXECUTE_AGAIN;
      }
      catch (IOException e)
      {
        logger.log(BasicLevel.WARN, e);
        return EXECUTE_AGAIN;
      }
      try
      {
        setSocketOption(socket);
      }
      catch (SocketException e)
      {
        logger.log(BasicLevel.WARN, e);
        try
        {
          socket.close();
        }
        catch (IOException e1)
        {
          // ignore.
        }
        return EXECUTE_AGAIN;
      }
      logger.log(BasicLevel.DEBUG, "ChannelIn-connection-task : connected");

      synchronized (availableSocketList)
      {
        nbConnection++;
        availableSocketList.add(new ConnectionState(socket));
        try
        {
          logger.log(BasicLevel.DEBUG,
              "ChannelIn-connection-task : add a thread in thread pool");
          threadPoolController.addThreads(1);
        }
        catch (ThreadPoolOverflowException e)
        {
          logger.log(BasicLevel.ERROR,
              "ChannelIn-connection-task : Unable to add reader thread", e);
          return STOP_EXECUTING;
        }
        catch (IllegalLifeCycleException e)
        {
          Error.bug(logger, e);
        }
        // if maxConnection reach, wait.
        while (nbConnection >= maxConnection)
        {
          availableSocketList.wait();
        }
      }
      return EXECUTE_AGAIN;
    }
  }

  protected class InTask extends AbstractTask
  {

    /** Constructor */
    public InTask()
    {
      super("ChannelIn-reader-task");
    }

    /**
     * @see Task#execute(Object)
     */
    public Object execute(Object hints) throws InterruptedException
    {
      ConnectionState connectionState;
      synchronized (availableSocketList)
      {
        connectionState = (ConnectionState) availableSocketList.removeFirst();
      }
      logger.log(BasicLevel.DEBUG, "ChannelIn-reader-task : got a connection");
      if (connectionState.isClosed())
      {
        stopThread(connectionState);
        logger.log(BasicLevel.DEBUG,
            "ChannelIn-reader-task : connection closed, stop thread");
        return STOP_EXECUTING;
      }
      try
      {

        ObjectInput input = connectionState.getInput();
        Message msg = messageCodecItf.decode(input);

        logger
            .log(BasicLevel.DEBUG, "ChannelIn-reader-task : message received");
        outPushItf.push(msg, null);
        // then send the acknowledge.
        logger.log(BasicLevel.DEBUG,
            "ChannelIn-reader-task : message pushed, send ack");

        sendAck(connectionState.socket);
      }
      catch (IOException e)
      {
        logger
            .log(
                BasicLevel.WARN,
                "ChannelIn-reader-task : I/O error while receiving message, close connection",
                e);
        stopThread(connectionState);
        return STOP_EXECUTING;
      }
      catch (PushException e)
      {
        logger.log(BasicLevel.WARN, e);
      }
      synchronized (availableSocketList)
      {
        availableSocketList.add(connectionState);
      }
      return EXECUTE_AGAIN;
    }

    protected void stopThread(ConnectionState connectionState)
    {
      synchronized (availableSocketList)
      {
        nbConnection--;
        // notify connection task.
        availableSocketList.notify();
      }
      connectionState.close();
    }
  }

  // ---------------------------------------------------------------------------
  // Overriding of initComponent method
  // ---------------------------------------------------------------------------

  protected void initComponent(Component componentItf)
      throws InitializationException
  {
    super.initComponent(componentItf);
    try
    {
      Util.addTask(componentItf, connectionTask, Collections.EMPTY_MAP);
      Map hints = new HashMap();
      hints.put("thread", "pool");
      logger.log(BasicLevel.DEBUG, "Initial Max connection=" + maxConnection);
      Util.addTask(componentItf, inTask, hints);
    }
    catch (Exception e)
    {
      throw new InitializationException(e, componentItf, "Can't add task");
    }
  }

  // ---------------------------------------------------------------------------
  // Overriden AbstractTCPChannelInImpl method
  // ---------------------------------------------------------------------------

  /**
   * @see AbstractTCPChannelInImpl#setSocketOption(Socket)
   */
  protected void setSocketOption(Socket sock) throws SocketException
  {
    if (soTimeout >= 0)
    {
      sock.setSoTimeout(soTimeout);
    }
  }

  // ---------------------------------------------------------------------------
  // Implementation of LifeCycleController interface
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.dream.AbstractComponent#startFc()
   */
  public void startFc() throws IllegalLifeCycleException
  {
    super.startFc();
    try
    {
      TaskController tc = (TaskController) weaveableC
          .getFcInterface("task-controller");
      threadPoolController = (ThreadPoolController) tc.getTaskControl(inTask);
      threadPoolAttributeController = (ThreadPoolAttributeController) Fractal
          .getAttributeController(((Interface) threadPoolController)
              .getFcItfOwner());
      threadPoolAttributeController.setCapacity(getMaxConnection());
    }
    catch (Exception e)
    {
      throw new ChainedIllegalLifeCycleException(e, null,
          "An error occurs while retreiving task control interface");
    }

  }

  /**
   * @see org.objectweb.fractal.api.control.LifeCycleController#stopFc()
   */
  public void stopFc() throws IllegalLifeCycleException
  {
    super.stopFc();
    threadPoolController = null;
    threadPoolAttributeController = null;
  }

  // ---------------------------------------------------------------------------
  // Implementation of MultiConnTCPChannelInAttributeController interface
  // ---------------------------------------------------------------------------

  /**
   * @see MultiConnTCPChannelInAttributeController#getMaxConnection()
   */
  public int getMaxConnection()
  {
    return maxConnection;
  }

  /**
   * @see MultiConnTCPChannelInAttributeController#setMaxConnection(int)
   */
  public void setMaxConnection(int maxConnection)
  {
    this.maxConnection = maxConnection;
    if (threadPoolAttributeController != null)
    {
      threadPoolAttributeController.setCapacity(maxConnection);
    }
  }

  /**
   * @see MultiConnTCPChannelInAttributeController#getSoTimeout()
   */
  public int getSoTimeout()
  {
    return soTimeout;
  }

  /**
   * @see MultiConnTCPChannelInAttributeController#setSoTimeout(int)
   */
  public void setSoTimeout(int timeout)
  {
    soTimeout = timeout;
  }

  // ---------------------------------------------------------------------------
  // Utility inner class
  // ---------------------------------------------------------------------------

  protected class ConnectionState
  {
    protected Socket             socket;
    protected ObjectInputStream  input;
    protected ObjectOutputStream output;

    protected ConnectionState(Socket socket)
    {
      this.socket = socket;
    }

    protected boolean isClosed()
    {
      return socket.isClosed() || !socket.isConnected();
    }

    protected ObjectInput getInput() throws IOException
    {
      if (input == null)
      {
        input = new ObjectInputStream(socket.getInputStream());
      }
      return input;
    }

    protected ObjectOutput getOutput() throws IOException
    {
      if (output == null)
      {
        output = new ObjectOutputStream(socket.getOutputStream());
      }
      return output;
    }

    protected void close()
    {
      if (input != null)
      {
        try
        {
          input.close();
        }
        catch (IOException e)
        {
          // ignore
        }
      }
      if (output != null)
      {
        try
        {
          output.close();
        }
        catch (IOException e)
        {
          // ignore
        }
      }
      if (!socket.isClosed())
      {
        try
        {
          socket.close();
        }
        catch (IOException e)
        {
          // ignore
        }
      }
    }
  }
}