/**
 * DREAM
 * Copyright (C) 2003 2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact : dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s):
 */

package org.objectweb.dream.protocol.causality;

/**
 * Component interfae to manipulate causal matrix clock. The causality protocol
 * composite allows <i>n </i> processes to exchange causaly ordered messages.
 * Each process has a identifier form <i>0 </i> to <i>n-1 </i>. When a process
 * send a message to an other, this message is stamped by the causality
 * protocol. Before to be delivered to the destination process, the stamp of a
 * given message is analysed to know if other messages should be delivered
 * before this one.
 */
public interface MatrixClock
{

  /** The default name of {@link MatrixClock }interfaces */
  String MATRIX_CLOCK_ITF_NAME = "matrixClock";

  /**
   * the value returned by
   * {@link MatrixClock#testRecvMatrix(Object, short, short) }if the message
   * can be delivred
   */
  int    DELIVER               = 0;

  /**
   * the value returned by
   * {@link MatrixClock#testRecvMatrix(Object, short, short) }if there is other
   * message in the causal ordering before this.
   */
  int    WAIT_TO_DELIVER       = 1;

  /**
   * the value returned by
   * {@link MatrixClock#testRecvMatrix(Object, short, short) }if the message
   * has already been delivred.
   */
  int    ALREADY_DELIVERED     = 2;

  /**
   * Test if the received message, the given stamp belong, can be delivered. If
   * the message is ready to be delivered, the method returns
   * <code>DELIVER</code> and the matrix clock is updated. If the message has
   * already been delivered, the method returns <code>ALREADY_DELIVERED</code>,
   * and if other messages are waited before this message the method returns
   * <code>WAIT_TO_DELIVER</code>. In the last two case the matrix clock
   * remains unchanged.
   * 
   * @param stamp a stamp
   * @param from the identifier of the sender process.
   * @param to the identifier of the local process.
   * @return <code>DELIVER</code>,<code>ALREADY_DELIVERED</code>, or
   *         <code>WAIT_TO_DELIVER</code> code.
   */
  int testRecvMatrix(Object stamp, short from, short to);

  /**
   * Generate a stamp for the specified destination identifier.
   * 
   * @param from the local process identifier
   * @param to the destination process identifier
   * @return a stamp.
   */
  Object getStamp(short from, short to);

}