/**
 * DREAM
 * Copyright (C) 2003 2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): 
 */

package org.objectweb.dream.queue;

import java.util.Map;

import org.objectweb.dream.Push;
import org.objectweb.dream.PushException;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.manager.MessageManager;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * Push/Push queue implemented using a passive component. It use the
 * {@link org.objectweb.dream.queue.PushPushPassiveQueueAttributeController }
 * attribute controller to open or close its output.
 */
public class BasicPushPushPassiveQueueImpl extends AbstractPushQueueImpl
    implements
      PushPushPassiveQueueAttributeController
{

  protected Push    outPushItf;
  protected boolean open = true;

  /**
   * @see Push#push(Message, Map)
   */
  public synchronized void push(Message message, Map context)
      throws PushException
  {
    if (open)
    {
      outPushItf.push(message, null);
    }
    else
    {
      super.push(message, context);
    }
  }

  // ---------------------------------------------------------------------------
  // Implementation of PushPushPassiveQueueAttributeController interface
  // ---------------------------------------------------------------------------

  /**
   * @see PushPushPassiveQueueAttributeController#getOutputStatus()
   */
  public synchronized boolean getOutputStatus()
  {
    return open;
  }

  /**
   * @see PushPushPassiveQueueAttributeController#setOutputStatus(boolean)
   */
  public synchronized void setOutputStatus(boolean outputStatus)
  {
    if (outputStatus == open)
    {
      // no change
      return;
    }
    open = outputStatus;
    if (outputStatus)
    {
      // output opened, push all waiting messages.
      while (!waitingList.isEmpty())
      {
        Message msg = (Message) waitingList.removeFirst();
        try
        {
          outPushItf.push(msg, null);
        }
        catch (PushException e)
        {
          logger.log(BasicLevel.INFO,
              "Exception catched while pushing a message, message dropped", e);
          messageManagerItf.deleteMessage(msg);
        }
      }
    }
  }

  // ---------------------------------------------------------------------------
  // Implementation of BindingController interface
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return new String[]{Push.OUT_PUSH_ITF_NAME, MessageManager.ITF_NAME};
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#bindFc(String,
   *      Object)
   */
  public synchronized void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    super.bindFc(clientItfName, serverItf);
    if (clientItfName.equals(Push.OUT_PUSH_ITF_NAME))
    {
      outPushItf = (Push) serverItf;
    }
  }
}